@@ -27,35 +27,38 @@
 
 # This test suite has been run successfully on the following systems:
 #
-# -------------------------------------------------------
+# -----------------------------------------------------------
 #  System                          | Python version
-# -------------------------------------------------------
-#  Linux CentOS 2.6.20.15          | 2.4, 2.5
+# -----------------------------------------------------------
 #  Linux Ubuntu 2.6.20-15          | 2.4, 2.5
+#  Linux Kubuntu 8.04 32 & 64 bits | 2.5.2
 #  Linux Debian 2.4.27-2-386       | 2.3.5
-#  Windows 2000 SP4                | 2.5
-#  Windows XP prof SP3             | 2.3, 2.4, 2.5, 2.6a3
-#  Windows Vista 32 & 64 bit       | 2.5
+#  Windows XP prof SP3             | 2.3, 2.4, 2.5, 2.6-RC2
+#  Windows Vista Ultimate 64 bit   | 2.5.1
+#  Windows Vista Business 32 bit   | 2.5.1
+#  Windows Server 2008 64bit       | 2.5.1
+#  Windows Mobile 6.1              | PythonCE 2.5
 #  OS X 10.4.10                    | 2.3, 2.4, 2.5
-#  FreeBSD 6.2, 7.0-RC1            | 2.4, 2.5
-#  Windows Mobile 6.0              | PythonCE 2.5
-# -------------------------------------------------------
+#  FreeBSD 7.0                     | 2.4, 2.5
+# -----------------------------------------------------------
 
 
 import threading
 import unittest
 import socket
 import os
+import time
 import re
 import tempfile
 import ftplib
 import random
 import warnings
+import sys
 
 from pyftpdlib import ftpserver
 
 
-__release__ = 'pyftpdlib 0.4.0'
+__release__ = 'pyftpdlib 0.5.0'
 
 # Attempt to use IP rather than hostname (test suite will run a lot faster)
 try:
@@ -86,7 +89,7 @@ SUPPORTS_IPV4 = try_address('127.0.0.1')
 SUPPORTS_IPV6 = socket.has_ipv6 and try_address('::1')
 
 
-class AbstractedFSClass(unittest.TestCase):
+class TestAbstractedFS(unittest.TestCase):
     """Test for conversion utility methods of AbstractedFS class."""
 
     def test_ftpnorm(self):
@@ -173,7 +176,7 @@ class AbstractedFSClass(unittest.TestCase):
         # Tests for fs2ftp method.
         ae = self.assertEquals
         fs = ftpserver.AbstractedFS()
-        join = lambda x,y: os.path.join(x, y.replace('/', os.sep))
+        join = lambda x, y: os.path.join(x, y.replace('/', os.sep))
 
         def goforit(root):
             fs.root = root
@@ -203,7 +206,7 @@ class AbstractedFSClass(unittest.TestCase):
             ae(fs.fs2ftp('D:\\dir'), '/')
         elif os.sep == '/':
             goforit('/')
-            assert os.path.realpath('/__home/user') == '/__home/user',\
+            assert os.path.realpath('/__home/user') == '/__home/user', \
                                     'Test skipped (symlinks not allowed).'
             goforit('/__home/user')
             fs.root = '/__home/user'
@@ -264,16 +267,37 @@ class AbstractedFSClass(unittest.TestCase):
                 file.close()
 
 
-class DummyAuthorizerClass(unittest.TestCase):
+class TestDummyAuthorizer(unittest.TestCase):
+    """Tests for DummyAuthorizer class."""
 
     # temporarily change warnings to exceptions for the purposes of testing
     def setUp(self):
+        self.tempdir = tempfile.mkdtemp(dir=HOME)
+        self.subtempdir = tempfile.mkdtemp(dir=os.path.join(HOME, self.tempdir))
+        self.tempfile = open(os.path.join(self.tempdir, TESTFN), 'w').name
+        self.subtempfile = open(os.path.join(self.subtempdir, TESTFN), 'w').name
         warnings.filterwarnings("error")
 
     def tearDown(self):
+        os.remove(self.tempfile)
+        os.remove(self.subtempfile)
+        os.rmdir(self.subtempdir)
+        os.rmdir(self.tempdir)
         warnings.resetwarnings()
 
-    def test_dummy_authorizer(self):
+    def assertRaisesWithMsg(self, excClass, msg, callableObj, *args, **kwargs):
+        try:
+            callableObj(*args, **kwargs)
+        except excClass, why:
+            if str(why) == msg:
+                return
+            raise self.failureException("%s != %s" %(str(why), msg))
+        else:
+            if hasattr(excClass,'__name__'): excName = excClass.__name__
+            else: excName = str(excClass)
+            raise self.failureException, "%s not raised" % excName
+
+    def test_common_methods(self):
         auth = ftpserver.DummyAuthorizer()
         # create user
         auth.add_user(USER, PASSWD, HOME)
@@ -284,33 +308,190 @@ class DummyAuthorizerClass(unittest.TestCase):
         # remove them
         auth.remove_user(USER)
         auth.remove_user('anonymous')
-
         # raise exc if user does not exists
         self.assertRaises(KeyError, auth.remove_user, USER)
         # raise exc if path does not exist
-        self.assertRaises(ftpserver.AuthorizerError, auth.add_user, USER,
-                            PASSWD, '?:\\')
-        self.assertRaises(ftpserver.AuthorizerError, auth.add_anonymous, '?:\\')
+        self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+                                'No such directory: "%s"' %'?:\\',
+                                 auth.add_user, USER, PASSWD, '?:\\')
+        self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+                                'No such directory: "%s"' %'?:\\',
+                                 auth.add_anonymous, '?:\\')
         # raise exc if user already exists
         auth.add_user(USER, PASSWD, HOME)
         auth.add_anonymous(HOME)
-        self.assertRaises(ftpserver.AuthorizerError, auth.add_user, USER,
-                            PASSWD, HOME)
-        self.assertRaises(ftpserver.AuthorizerError, auth.add_anonymous, HOME)
+        self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+                                'User "%s" already exists' %USER,
+                                 auth.add_user, USER, PASSWD, HOME)
+        self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+                                'User "anonymous" already exists',
+                                 auth.add_anonymous, HOME)
         auth.remove_user(USER)
         auth.remove_user('anonymous')
-
         # raise on wrong permission
-        self.assertRaises(ftpserver.AuthorizerError, auth.add_user, USER,
-                          PASSWD, HOME, perm='?')
-        self.assertRaises(ftpserver.AuthorizerError, auth.add_anonymous, HOME,
-                            perm='?')
+        self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+                                 'No such permission "?"',
+                                 auth.add_user, USER, PASSWD, HOME, perm='?')
+        self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+                                 'No such permission "?"',
+                                 auth.add_anonymous, HOME, perm='?')
         # expect warning on write permissions assigned to anonymous user
         for x in "adfmw":
-            self.assertRaises(RuntimeWarning, auth.add_anonymous, HOME, perm=x)
+            self.assertRaisesWithMsg(RuntimeWarning,
+                                "Write permissions assigned to anonymous user.",
+                                auth.add_anonymous, HOME, perm=x)
+
+    def test_override_perm_interface(self):
+        auth = ftpserver.DummyAuthorizer()
+        auth.add_user(USER, PASSWD, HOME, perm='elr')
+        # raise exc if user does not exists
+        self.assertRaises(KeyError, auth.override_perm, USER+'w', HOME, 'elr')
+        # raise exc if path does not exist or it's not a directory
+        self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+                                'No such directory: "%s"' %'?:\\',
+                                auth.override_perm, USER, '?:\\', 'elr')
+        self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+                                'No such directory: "%s"' %self.tempfile,
+                                auth.override_perm, USER, self.tempfile, 'elr')
+        # raise on wrong permission
+        self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+                                 'No such permission "?"', auth.override_perm,
+                                 USER, HOME, perm='?')
+        # expect warning on write permissions assigned to anonymous user
+        auth.add_anonymous(HOME)
+        for p in "adfmw":
+            self.assertRaisesWithMsg(RuntimeWarning,
+                                "Write permissions assigned to anonymous user.",
+                                auth.override_perm, 'anonymous', HOME, p)
+        # raise on attempt to override home directory permissions
+        self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+                                 "Can't override home directory permissions",
+                                 auth.override_perm, USER, HOME, perm='w')
+        # raise on attempt to override a path escaping home directory
+        if os.path.dirname(HOME) != HOME:
+            self.assertRaisesWithMsg(ftpserver.AuthorizerError,
+                                     "Path escapes user home directory",
+                                     auth.override_perm, USER,
+                                     os.path.dirname(HOME), perm='w')
+        # try to re-set an overridden permission
+        auth.override_perm(USER, self.tempdir, perm='w')
+        auth.override_perm(USER, self.tempdir, perm='wr')
+
+    def test_override_perm_recursive_paths(self):
+        auth = ftpserver.DummyAuthorizer()
+        auth.add_user(USER, PASSWD, HOME, perm='elr')
+        self.assert_(auth.has_perm(USER, 'w', self.tempdir) is False)
+        auth.override_perm(USER, self.tempdir, perm='w', recursive=True)
+        self.assert_(auth.has_perm(USER, 'w', HOME) is False)
+        self.assert_(auth.has_perm(USER, 'w', self.tempdir) is True)
+        self.assert_(auth.has_perm(USER, 'w', self.tempfile) is True)
+        self.assert_(auth.has_perm(USER, 'w', self.subtempdir) is True)
+        self.assert_(auth.has_perm(USER, 'w', self.subtempfile) is True)
+
+        self.assert_(auth.has_perm(USER, 'w', HOME + '@') is False)
+        self.assert_(auth.has_perm(USER, 'w', self.tempdir + '@') is False)
+        path = os.path.join(self.tempdir + '@', os.path.basename(self.tempfile))
+        self.assert_(auth.has_perm(USER, 'w', path) is False)
+        # test case-sensitiveness
+        if (os.name in ('nt', 'ce')) or (sys.platform == 'cygwin'):
+            self.assert_(auth.has_perm(USER, 'w', self.tempdir.upper()) is True)
+
+    def test_override_perm_not_recursive_paths(self):
+        auth = ftpserver.DummyAuthorizer()
+        auth.add_user(USER, PASSWD, HOME, perm='elr')
+        self.assert_(auth.has_perm(USER, 'w', self.tempdir) is False)
+        auth.override_perm(USER, self.tempdir, perm='w')
+        self.assert_(auth.has_perm(USER, 'w', HOME) is False)
+        self.assert_(auth.has_perm(USER, 'w', self.tempdir) is True)
+        self.assert_(auth.has_perm(USER, 'w', self.tempfile) is True)
+        self.assert_(auth.has_perm(USER, 'w', self.subtempdir) is False)
+        self.assert_(auth.has_perm(USER, 'w', self.subtempfile) is False)
+
+        self.assert_(auth.has_perm(USER, 'w', HOME + '@') is False)
+        self.assert_(auth.has_perm(USER, 'w', self.tempdir + '@') is False)
+        path = os.path.join(self.tempdir + '@', os.path.basename(self.tempfile))
+        self.assert_(auth.has_perm(USER, 'w', path) is False)
+        # test case-sensitiveness
+        if (os.name in ('nt', 'ce')) or (sys.platform == 'cygwin'):
+            self.assert_(auth.has_perm(USER, 'w', self.tempdir.upper()) is True)
 
 
-class FtpAuthentication(unittest.TestCase):
+class TestCallLater(unittest.TestCase):
+    """Tests for CallLater class."""
+
+    def setUp(self):
+        for task in ftpserver._tasks:
+            if not task.cancelled:
+                task.cancel()
+        del ftpserver._tasks[:]
+
+    def scheduler(self, timeout=0.01, count=100):
+        while ftpserver._tasks and count > 0:
+            ftpserver._scheduler()
+            count -= 1
+            time.sleep(timeout)
+
+    def test_interface(self):
+        fun = lambda: 0
+        self.assertRaises(AssertionError, ftpserver.CallLater, -1, fun)
+        x = ftpserver.CallLater(3, fun)
+        self.assertRaises(AssertionError, x.delay, -1)
+        self.assert_(x.cancelled is False)
+        x.cancel()
+        self.assert_(x.cancelled is True)
+        self.assertRaises(AssertionError, x.call)
+        self.assertRaises(AssertionError, x.reset)
+        self.assertRaises(AssertionError, x.delay, 2)
+        self.assertRaises(AssertionError, x.cancel)
+
+    def test_order(self):
+        l = []
+        fun = lambda x: l.append(x)
+        for x in [0.05, 0.04, 0.03, 0.02, 0.01]:
+            ftpserver.CallLater(x, fun, x)
+        self.scheduler()
+        self.assertEqual(l, [0.01, 0.02, 0.03, 0.04, 0.05])
+
+    def test_delay(self):
+        l = []
+        fun = lambda x: l.append(x)
+        ftpserver.CallLater(0.01, fun, 0.01).delay(0.07)
+        ftpserver.CallLater(0.02, fun, 0.02).delay(0.08)
+        ftpserver.CallLater(0.03, fun, 0.03)
+        ftpserver.CallLater(0.04, fun, 0.04)
+        ftpserver.CallLater(0.05, fun, 0.05)
+        ftpserver.CallLater(0.06, fun, 0.06).delay(0.001)
+        self.scheduler()
+        self.assertEqual(l, [0.06, 0.03, 0.04, 0.05, 0.01, 0.02])
+
+    def test_reset(self):
+        # will fail on such systems where time.time() does not provide
+        # time with a better precision than 1 second.
+        l = []
+        fun = lambda x: l.append(x)
+        ftpserver.CallLater(0.01, fun, 0.01)
+        ftpserver.CallLater(0.02, fun, 0.02)
+        ftpserver.CallLater(0.03, fun, 0.03)
+        x = ftpserver.CallLater(0.04, fun, 0.04)
+        ftpserver.CallLater(0.05, fun, 0.05)
+        time.sleep(0.1)
+        x.reset()
+        self.scheduler()
+        self.assertEqual(l, [0.01, 0.02, 0.03, 0.05, 0.04])
+
+    def test_cancel(self):
+        l = []
+        fun = lambda x: l.append(x)
+        ftpserver.CallLater(0.01, fun, 0.01).cancel()
+        ftpserver.CallLater(0.02, fun, 0.02)
+        ftpserver.CallLater(0.03, fun, 0.03)
+        ftpserver.CallLater(0.04, fun, 0.04)
+        ftpserver.CallLater(0.05, fun, 0.05).cancel()
+        self.scheduler()
+        self.assertEqual(l, [0.02, 0.03, 0.04])
+
+
+class TestFtpAuthentication(unittest.TestCase):
     "test: USER, PASS, REIN."
 
     def setUp(self):
@@ -334,25 +515,29 @@ class FtpAuthentication(unittest.TestCase):
     def test_auth_ok(self):
         self.client.login(user=USER, passwd=PASSWD)
 
-    def test_auth_failed(self):
-        self.assertRaises(ftplib.error_perm, self.client.login, USER,
-                              passwd='wrong')
-
     def test_anon_auth(self):
         self.client.login(user='anonymous', passwd='anon@')
         self.client.login(user='AnonYmoUs', passwd='anon@')
         self.client.login(user='anonymous', passwd='')
 
-    def test_max_auth(self):
-        self.assertRaises(ftplib.error_perm, self.client.login, USER, 'wrong')
-        self.assertRaises(ftplib.error_perm, self.client.login, USER, 'wrong')
-        self.assertRaises(ftplib.error_perm, self.client.login, USER, 'wrong')
-        # If authentication fails for 3 times ftpd disconnects the
-        # client.  We can check if that happens by using self.client.sendcmd()
-        # on the 'dead' socket object.  If socket object is really
-        # closed it should be raised a socket.error exception (Windows)
-        # or a EOFError exception (Linux).
-        self.assertRaises((socket.error, EOFError), self.client.sendcmd, '')
+    # Commented after delayed response on wrong credentials has been
+    # introduced because tests take too much to complete.
+
+##    def test_auth_failed(self):
+##        self.assertRaises(ftplib.error_perm, self.client.login, USER, 'wrong')
+##        self.assertRaises(ftplib.error_perm, self.client.login, 'wrong', PASSWD)
+##        self.assertRaises(ftplib.error_perm, self.client.login, 'wrong', 'wrong')
+
+##    def test_max_auth(self):
+##        self.assertRaises(ftplib.error_perm, self.client.login, USER, 'wrong')
+##        self.assertRaises(ftplib.error_perm, self.client.login, USER, 'wrong')
+##        self.assertRaises(ftplib.error_perm, self.client.login, USER, 'wrong')
+##        # If authentication fails for 3 times ftpd disconnects the
+##        # client.  We can check if that happens by using self.client.sendcmd()
+##        # on the 'dead' socket object.  If socket object is really
+##        # closed it should be raised a socket.error exception (Windows)
+##        # or a EOFError exception (Linux).
+##        self.assertRaises((socket.error, EOFError), self.client.sendcmd, '')
 
     def test_rein(self):
         self.client.login(user=USER, passwd=PASSWD)
@@ -372,21 +557,17 @@ class FtpAuthentication(unittest.TestCase):
 
         self.client.voidcmd('TYPE I')
         conn = self.client.transfercmd('retr ' + TESTFN)
-        bytes_recv = 0
         rein_sent = 0
         while 1:
             chunk = conn.recv(8192)
-            # stop transfer while it isn't finished yet
-            if bytes_recv >= 524288: # 2^19
-                if not rein_sent:
-                    # flush account, error response expected
-                    self.client.sendcmd('rein')
-                    self.assertRaises(ftplib.error_perm, self.client.dir)
-                    rein_sent = 1
             if not chunk:
                 break
             self.f2.write(chunk)
-            bytes_recv += len(chunk)
+            if not rein_sent:
+                rein_sent = 1
+                # flush account, error response expected
+                self.client.sendcmd('rein')
+                self.assertRaises(ftplib.error_perm, self.client.dir)
 
         # a 226 response is expected once tranfer finishes
         self.assertEqual(self.client.voidresp()[:3], '226')
@@ -419,21 +600,18 @@ class FtpAuthentication(unittest.TestCase):
 
         self.client.voidcmd('TYPE I')
         conn = self.client.transfercmd('retr ' + TESTFN)
-        bytes_recv = 0
         rein_sent = 0
         while 1:
             chunk = conn.recv(8192)
-            # stop transfer while it isn't finished yet
-            if bytes_recv >= 524288: # 2^19
-                if not rein_sent:
-                    # flush account, expect an error response
-                    self.client.sendcmd('user ' + USER)
-                    self.assertRaises(ftplib.error_perm, self.client.dir)
-                    rein_sent = 1
             if not chunk:
                 break
             self.f2.write(chunk)
-            bytes_recv += len(chunk)
+            # stop transfer while it isn't finished yet
+            if not rein_sent:
+                rein_sent = 1
+                # flush account, expect an error response
+                self.client.sendcmd('user ' + USER)
+                self.assertRaises(ftplib.error_perm, self.client.dir)
 
         # a 226 response is expected once tranfer finishes
         self.assertEqual(self.client.voidresp()[:3], '226')
@@ -447,7 +625,7 @@ class FtpAuthentication(unittest.TestCase):
         self.assertEqual(hash(data), hash (self.f2.read()))
 
 
-class FtpDummyCmds(unittest.TestCase):
+class TestFtpDummyCmds(unittest.TestCase):
     "test: TYPE, STRU, MODE, NOOP, SYST, ALLO, HELP"
 
     def setUp(self):
@@ -464,19 +642,16 @@ class FtpDummyCmds(unittest.TestCase):
     def test_type(self):
         self.client.sendcmd('type a')
         self.client.sendcmd('type i')
-        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'type')
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'type ?!?')
 
     def test_stru(self):
         self.client.sendcmd('stru f')
         self.client.sendcmd('stru F')
-        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'stru')
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'stru ?!?')
 
     def test_mode(self):
         self.client.sendcmd('mode s')
         self.client.sendcmd('mode S')
-        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'mode')
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'mode ?!?')
 
     def test_noop(self):
@@ -488,6 +663,9 @@ class FtpDummyCmds(unittest.TestCase):
     def test_allo(self):
         self.client.sendcmd('allo x')
 
+    def test_quit(self):
+        self.client.sendcmd('quit')
+
     def test_help(self):
         self.client.sendcmd('help')
         cmd = random.choice(ftpserver.proto_cmds.keys())
@@ -502,7 +680,6 @@ class FtpDummyCmds(unittest.TestCase):
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'rest -1')
 
     def test_opts_feat(self):
-        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'opts')
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'opts mlst bad_fact')
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'opts mlst type ;')
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'opts not_mlst')
@@ -527,11 +704,67 @@ class FtpDummyCmds(unittest.TestCase):
                                      '200 MLST OPTS type;')
         self.failUnless('type*;perm;size;modify;' in mlst())
 
-    def test_quit(self):
+
+class TestFtpCmdsSemantic(unittest.TestCase):
+
+    arg_cmds = ('allo','appe','dele','eprt','mdtm','mode','mkd','opts','port',
+                'rest','retr','rmd','rnfr','rnto','size', 'stor', 'stru','type',
+                'user','xmkd','xrmd')
+
+    def setUp(self):
+        self.server = FTPd()
+        self.server.start()
+        self.client = ftplib.FTP()
+        self.client.connect(self.server.host, self.server.port)
+        self.client.login(USER, PASSWD)
+
+    def tearDown(self):
+        self.client.close()
+        self.server.stop()
+
+    def test_arg_cmds(self):
+        # test commands requiring an argument
+        expected = "501 Syntax error: command needs an argument."
+        for cmd in self.arg_cmds:
+            self.client.putcmd(cmd)
+            resp = self.client.getmultiline()
+            self.assertEqual(resp, expected)
+
+    def test_no_arg_cmds(self):
+        # test commands accepting no arguments
+        expected = "501 Syntax error: command does not accept arguments."
+        for cmd in ('abor','cdup','feat','noop','pasv','pwd','quit','rein',
+                    'syst','xcup','xpwd'):
+            self.client.putcmd(cmd + ' arg')
+            resp = self.client.getmultiline()
+            self.assertEqual(resp, expected)
+
+    def test_auth_cmds(self):
+        # test those commands requiring client to be authenticated
+        expected = "530 Log in with USER and PASS first."
+        self.client.sendcmd('rein')
+        for cmd in ftpserver.proto_cmds:
+            cmd = cmd.lower()
+            if cmd in ('feat','help','noop','user','pass','stat','syst','quit'):
+                continue
+            if cmd in self.arg_cmds:
+                cmd = cmd + ' arg'
+            self.client.putcmd(cmd)
+            resp = self.client.getmultiline()
+            self.assertEqual(resp, expected)
+
+    def test_no_auth_cmds(self):
+        # test those commands that do not require client to be authenticated
+        self.client.sendcmd('rein')
+        for cmd in ('feat','help','noop','stat','syst'):
+            self.client.sendcmd(cmd)
+        # STAT provided with an argument is equal to LIST hence not allowed
+        # if not authenticated
+        self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'stat /')
         self.client.sendcmd('quit')
 
 
-class FtpFsOperations(unittest.TestCase):
+class TestFtpFsOperations(unittest.TestCase):
     "test: PWD, CWD, CDUP, SIZE, RNFR, RNTO, DELE, MKD, RMD, MDTM, STAT"
 
     def setUp(self):
@@ -554,7 +787,12 @@ class FtpFsOperations(unittest.TestCase):
 
     def test_cwd(self):
         self.client.cwd(self.tempdir)
+        self.assertEqual(self.client.pwd(), '/' + self.tempdir)
         self.assertRaises(ftplib.error_perm, self.client.cwd, 'subtempdir')
+        # cwd provided with no arguments is supposed to move us to the
+        # root directory
+        self.client.sendcmd('cwd')
+        self.assertEqual(self.client.pwd(), '/')
 
     def test_pwd(self):
         self.assertEqual(self.client.pwd(), '/')
@@ -562,16 +800,13 @@ class FtpFsOperations(unittest.TestCase):
         self.assertEqual(self.client.pwd(), '/' + self.tempdir)
 
     def test_cdup(self):
-        # ftplib.parse257 function is usually used for parsing the
-        # '257' response for a MKD or PWD request returning the
-        # directory name in the 257 reply.
-        # Although CDUP response code is different (250) we can use
-        # parse257 for getting directory name.
         self.client.cwd(self.tempdir)
-        dir = ftplib.parse257(self.client.sendcmd('cdup').replace('250', '257'))
-        self.assertEqual(dir, '/')
-        dir = ftplib.parse257(self.client.sendcmd('cdup').replace('250', '257'))
-        self.assertEqual(dir, '/')
+        self.assertEqual(self.client.pwd(), '/' + self.tempdir)
+        self.client.sendcmd('cdup')
+        self.assertEqual(self.client.pwd(), '/')
+        # make sure we can't escape from root directory
+        self.client.sendcmd('cdup')
+        self.assertEqual(self.client.pwd(), '/')
 
     def test_mkd(self):
         tempdir = os.path.basename(tempfile.mktemp(dir=HOME))
@@ -633,11 +868,8 @@ class FtpFsOperations(unittest.TestCase):
         else:
             self.fail('Exception not raised')
 
-    def test_stat(self):
-        self.client.sendcmd('stat')
 
-
-class FtpRetrieveData(unittest.TestCase):
+class TestFtpRetrieveData(unittest.TestCase):
     "test: RETR, REST, LIST, NLST, argumented STAT"
 
     def setUp(self):
@@ -670,25 +902,15 @@ class FtpRetrieveData(unittest.TestCase):
     def test_restore_on_retr(self):
         data = 'abcde12345' * 100000
         fname_1 = os.path.basename(self.f1.name)
-        fname_2 = os.path.basename(self.f2.name)
         self.f1.write(data)
         self.f1.close()
 
         # look at ftplib.FTP.retrbinary method to understand this mess
         self.client.voidcmd('TYPE I')
         conn = self.client.transfercmd('retr ' + fname_1)
-        bytes_recv = 0
-        while 1:
-            chunk = conn.recv(8192)
-            # stop transfer while it isn't finished yet
-            if bytes_recv >= 524288: # 2^19
-                break
-            elif not chunk:
-                break
-            self.f2.write(chunk)
-            bytes_recv += len(chunk)
+        chunk = conn.recv(len(data) / 2)
+        self.f2.write(chunk)
         conn.close()
-
         # transfer wasn't finished yet so we expect a 426 response
         self.assertRaises(ftplib.error_temp, self.client.voidresp)
 
@@ -700,7 +922,7 @@ class FtpRetrieveData(unittest.TestCase):
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'retr ' + fname_1)
 
         # test resume
-        self.client.sendcmd('rest %s' %bytes_recv)
+        self.client.sendcmd('rest %s' %len(chunk))
         self.client.retrbinary("retr " + fname_1, self.f2.write)
         self.f2.seek(0)
         self.assertEqual(hash(data), hash (self.f2.read()))
@@ -792,16 +1014,17 @@ class FtpRetrieveData(unittest.TestCase):
             os.rmdir(dir)
 
     def test_stat(self):
-        # test argumented STAT which is equal to LIST plus globbing support
-        self.client.sendcmd('stat *')
+        # test STAT provided with argument which is equal to LIST
         self.client.sendcmd('stat /')
         self.client.sendcmd('stat ' + TESTFN)
-        self.failUnless('recursion not supported' in self.client.sendcmd('stat /*/*'))
+        self.client.putcmd('stat *')
+        resp = self.client.getmultiline()
+        self.assertEqual(resp, '550 Globbing not supported.')
         bogus = os.path.basename(tempfile.mktemp(dir=HOME))
         self.assertRaises(ftplib.error_perm, self.client.sendcmd, 'stat ' + bogus)
 
 
-class FtpAbort(unittest.TestCase):
+class TestFtpAbort(unittest.TestCase):
     "test: ABOR"
 
     def setUp(self):
@@ -857,16 +1080,8 @@ class FtpAbort(unittest.TestCase):
         # in a separate thread
         self.client.voidcmd('TYPE I')
         conn = self.client.transfercmd('retr ' + TESTFN)
-        bytes_recv = 0
-        while 1:
-            chunk = conn.recv(8192)
-            # stop transfer while it isn't finished yet
-            if bytes_recv >= 524288: # 2^19
-                break
-            elif not chunk:
-                break
-            self.f2.write(chunk)
-            bytes_recv += len(chunk)
+        chunk = conn.recv(len(data) / 2)
+        # stop transfer while it isn't finished yet
         self.client.putcmd('ABOR')
 
         # transfer isn't finished yet so ftpd should respond with 426
@@ -875,8 +1090,22 @@ class FtpAbort(unittest.TestCase):
         # transfer successfully aborted, so should now respond with a 226
         self.failUnlessEqual('226', self.client.voidresp()[:3])
 
+    if hasattr(socket, 'MSG_OOB'):
+        def test_oob_abor(self):
+            # Send ABOR by following the RFC-959 directives of sending
+            # Telnet IP/Synch sequence as OOB data.
+            # On some systems like FreeBSD this happened to be a problem
+            # due to a different SO_OOBINLINE behavior.
+            # On some platforms (e.g. Python CE) the test may fail
+            # although the MSG_OOB constant is defined.
+            self.client.sock.sendall(chr(244), socket.MSG_OOB)
+            self.client.sock.sendall(chr(242), socket.MSG_OOB)
+            self.client.sock.sendall('abor\r\n')
+            self.client.sock.settimeout(1)
+            self.assertEqual(self.client.getresp()[:3], '225')
 
-class FtpStoreData(unittest.TestCase):
+
+class TestFtpStoreData(unittest.TestCase):
     "test: STOR, STOU, APPE, REST"
 
     def setUp(self):
@@ -1014,6 +1243,141 @@ class FtpStoreData(unittest.TestCase):
         self.client.delete(TESTFN3)
 
 
+class TestTimeouts(unittest.TestCase):
+    """Test idle-timeout capabilities of control and data channels.
+    Some tests may fail on slow machines.
+    """
+
+    def _setUp(self, idle_timeout=300, data_timeout=300, pasv_timeout=30):
+        self.server = FTPd()
+        self.server.handler.timeout = idle_timeout
+        self.server.handler.dtp_handler.timeout = data_timeout
+        self.server.handler.passive_dtp.timeout = pasv_timeout
+        self.server.start()
+        self.client = ftplib.FTP()
+        self.client.connect(self.server.host, self.server.port)
+        self.client.login(USER, PASSWD)
+
+    def tearDown(self):
+        self.client.close()
+        self.server.handler.timeout = 300
+        self.server.handler.dtp_handler.timeout = 300
+        self.server.handler.passive_dtp.timeout = 30
+        self.server.stop()
+
+    def test_idle_timeout(self):
+        # Test control channel timeout. The client which does not send
+        # any command within the time specified in FTPHandler.timeout is
+        # supposed to be kicked off.
+        self._setUp(idle_timeout=0.1)
+        # fail if no msg is received within 1 second
+        self.client.sock.settimeout(1)
+        data = self.client.sock.recv(1024)
+        self.assertEqual(data, "421 Control connection timed out.\r\n")
+        # ensure client has been kicked off
+        self.assertRaises((socket.error, EOFError), self.client.sendcmd, 'noop')
+
+    def test_data_timeout(self):
+        # Test data channel timeout. The client which does not send
+        # or receive any data within the time specified in
+        # DTPHandler.timeout is supposed to be kicked off.
+        self._setUp(data_timeout=0.1)
+        addr = self.client.makepasv()
+        s = socket.socket()
+        s.connect(addr)
+        # fail if no msg is received within 1 second
+        self.client.sock.settimeout(1)
+        data = self.client.sock.recv(1024)
+        self.assertEqual(data, "421 Data connection timed out.\r\n")
+        # ensure client has been kicked off
+        self.assertRaises((socket.error, EOFError), self.client.sendcmd, 'noop')
+
+    def test_idle_data_timeout1(self):
+        # Tests that the control connection timeout is suspended while
+        # the data channel is opened
+        self._setUp(idle_timeout=0.1, data_timeout=0.2)
+        addr = self.client.makepasv()
+        s = socket.socket()
+        s.connect(addr)
+        # fail if no msg is received within 1 second
+        self.client.sock.settimeout(1)
+        data = self.client.sock.recv(1024)
+        self.assertEqual(data, "421 Data connection timed out.\r\n")
+        # ensure client has been kicked off
+        self.assertRaises((socket.error, EOFError), self.client.sendcmd, 'noop')
+
+    def test_idle_data_timeout2(self):
+        # Tests that the control connection timeout is restarted after
+        # data channel has been closed
+        self._setUp(idle_timeout=0.1, data_timeout=0.2)
+        addr = self.client.makepasv()
+        s = socket.socket()
+        s.connect(addr)
+        # close data channel
+        self.client.sendcmd('abor')
+        self.client.sock.settimeout(1)
+        data = self.client.sock.recv(1024)
+        self.assertEqual(data, "421 Control connection timed out.\r\n")
+        # ensure client has been kicked off
+        self.assertRaises((socket.error, EOFError), self.client.sendcmd, 'noop')
+
+    def test_pasv_timeout(self):
+        # Test pasv data channel timeout. The client which does not connect
+        # to the listening data socket within the time specified in
+        # PassiveDTP.timeout is supposed to receive a 421 response.
+        self._setUp(pasv_timeout=0.1)
+        self.client.makepasv()
+        # fail if no msg is received within 1 second
+        self.client.sock.settimeout(1)
+        data = self.client.sock.recv(1024)
+        self.assertEqual(data, "421 Passive data channel timed out.\r\n")
+        # client is not expected to be kicked off
+        self.client.sendcmd('noop')
+
+
+class TestMaxConnections(unittest.TestCase):
+    """Test maximum connections (FTPServer.max_cons)."""
+
+    def setUp(self):
+        self.server = FTPd()
+        self.server.server.max_cons = 3
+        self.server.start()
+
+    def tearDown(self):
+        self.server.server.max_cons = 0
+        self.server.stop()
+
+    def test_max_connections(self):
+        c1 = ftplib.FTP()
+        c2 = ftplib.FTP()
+        c3 = ftplib.FTP()
+        try:
+            c1.connect(self.server.host, self.server.port)
+            c2.connect(self.server.host, self.server.port)
+            self.assertRaises(ftplib.error_temp, c3.connect, self.server.host,
+                              self.server.port)
+            # with passive data channel established
+            c2.close()
+            c1.login(USER, PASSWD)
+            c1.makepasv()
+            self.assertRaises(ftplib.error_temp, c2.connect, self.server.host,
+                              self.server.port)
+            # with passive data socket waiting for connection
+            c1.login(USER, PASSWD)
+            c1.sendcmd('pasv')
+            self.assertRaises(ftplib.error_temp, c2.connect, self.server.host,
+                              self.server.port)
+            # with active data channel established
+            c1.login(USER, PASSWD)
+            c1.makeport()
+            self.assertRaises(ftplib.error_temp, c2.connect, self.server.host,
+                              self.server.port)
+        finally:
+            c1.close()
+            c2.close()
+            c3.close()
+
+
 class _TestNetworkProtocols(unittest.TestCase):
     """Test PASV, EPSV, PORT and EPRT commands.
 
@@ -1204,22 +1568,13 @@ class FTPd(threading.Thread):
         self.__flag.wait()
 
     def run(self):
-        # hack for granting backward compatibility with Python 2.3
-        # where asyncore.loop() doesn't provide the 'count' argument
-        import inspect, asyncore
-        if 'count' in inspect.getargspec(asyncore.loop)[0]:
-            poller = self.server.serve_forever
-            kwargs = {"timeout":0.001, "count":1}
-        else:
-            poller = asyncore.poll
-            kwargs = {"timeout":0.001}
         self.active = True
         self.__flag.set()
         while self.active:
             self.active_lock.acquire()
-            poller(**kwargs)
+            self.server.serve_forever(timeout=0.001, count=1)
             self.active_lock.release()
-        self.server.close()
+        self.server.close_all(ignore_all=True)
 
     def stop(self):
         assert self.active
@@ -1239,14 +1594,18 @@ def test_main(tests=None):
     test_suite = unittest.TestSuite()
     if tests is None:
         tests = [
-                 AbstractedFSClass,
-                 DummyAuthorizerClass,
-                 FtpAuthentication,
-                 FtpDummyCmds,
-                 FtpFsOperations,
-                 FtpRetrieveData,
-                 FtpAbort,
-                 FtpStoreData
+                 TestAbstractedFS,
+                 TestDummyAuthorizer,
+                 TestCallLater,
+                 TestFtpAuthentication,
+                 TestFtpDummyCmds,
+                 TestFtpCmdsSemantic,
+                 TestFtpFsOperations,
+                 TestFtpRetrieveData,
+                 TestFtpAbort,
+                 TestFtpStoreData,
+                 TestTimeouts,
+                 TestMaxConnections
                  ]
         if SUPPORTS_IPV4:
             tests.append(TestIPv4Environment)
